#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-08-05
# File  : vzlogger.py
#
# vzlogger
#
#
# https://wiki.volkszaehler.org/software/controller/vzlogger
#
#
# 2023-08-05: initial release

import json
import time
from dataclasses import dataclass
from typing import Dict, Mapping, Tuple, Optional, Any
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    DiscoveryResult,
    CheckResult,
    StringTable,
)
from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    Service,
    Result,
    State,
    check_levels_predictive,
    check_levels,
)


@dataclass
class VzLoggerChannel:
    uuid: str
    interval: int
    last: int
    protocol: str
    value: float
    value_last: int


_vzlogger_channel_types = {
    'meter_reading': 'Meter reading',
    'voltage': 'Voltage',
    'electrical_power': 'Electrical Power',
}


def _render_voltage(value: float) -> str:
    return f'{value:.2f} V'


def _render_electrical_power(value: float) -> str:
    return f'{value:.2f} W'


def _render_meter_reading(value: float) -> str:
    return f'{value/1000:.0f} kWh'


def _yield_channel_details(channel: VzLoggerChannel) -> CheckResult:
    yield Result(state=State.OK, notice=f'UUID: {channel.uuid}')
    yield Result(state=State.OK, notice=f'Interval: {channel.interval}')
    yield Result(state=State.OK, notice=f'Last: {time.ctime(channel.last)}')
    yield Result(state=State.OK, notice=f'Protocol: {channel.protocol}')
    yield Result(state=State.OK, notice=f'Value: {channel.value}')
    yield Result(state=State.OK, notice=f'Value last: {time.ctime(channel.value_last)}')


def _yield_channel_value(
        value: float,
        label: str,
        metric: str,
        render_func: Any,
        levels_upper: Optional[Tuple[int, int]] = None,
        level_lower: Optional[Tuple[int, int]] = None,
) -> CheckResult:
    metric = metric.lower().replace(' ', '_').replace('-', '_')
    yield from check_levels_predictive(
        value=value,
        label=label,
        metric_name=f'{metric}',
        render_func=render_func,
        levels=levels_upper,
        boundaries=(0, None),
    ) if isinstance(levels_upper, dict) else check_levels(
        value=value,
        # label=label,
        metric_name=f'{metric}',
        render_func=render_func,
        levels_upper=levels_upper,
        levels_lower=level_lower,
        # boundaries=(0, None),
    )


def parse_vzlogger(string_table: StringTable) -> Dict[str, VzLoggerChannel]:
    section = {}
    data = json.loads(string_table[0][0])
    channels = data['data']
    for channel in channels:
        value_last, value = channel['tuples'][0]
        section[channel['uuid']] = VzLoggerChannel(
            uuid=channel['uuid'],
            interval=channel['interval'],
            last=channel['last'],
            protocol=channel['protocol'],
            value=value,
            value_last=value_last,
        )

    return section


def discovery_vzlogger(params: Mapping[str, any], section: Dict[str, VzLoggerChannel]) -> DiscoveryResult:
    # add vzlogger channels with info from discovery rule vzlogger
    for uuid, channel_type, item in params['channels']:
        if uuid in section.keys():
            yield Service(
                item=f'{_vzlogger_channel_types.get(channel_type, uuid)} {item}',
                parameters={
                    'channel_type': channel_type,
                    'uuid': uuid,
                }
            )
            section.pop(uuid)

    # add unknown vzlogger channels
    for uuid in section.keys():
        yield Service(
            item=f'vzlogger {uuid}',
            parameters={
                'channel_type': 'unknown',
                'uuid': uuid,
            }
        )


def check_vzlogger(item: str, params: Mapping[str, any], section: Dict[str, VzLoggerChannel]) -> CheckResult:
    try:
        channel = section[params['uuid']]
    except KeyError:
        yield Result(state=State.UNKNOWN, summary=f'Data not found in section')
        return

    if params['channel_type'] == 'voltage':
        yield from _yield_channel_value(
            value=channel.value,
            label=item[8:],
            metric=params['channel_type'],
            render_func=_render_voltage,
            levels_upper=params.get('levels_upper'),
            level_lower=params.get('levels_lower'),
        )

    elif params['channel_type'] == 'electrical_power':

        yield from _yield_channel_value(
            value=channel.value,
            label=item[17:],
            metric='power',
            render_func=_render_electrical_power,
            levels_upper=params.get('levels_upper'),
            level_lower=params.get('levels_lower'),
        )

    elif params['channel_type'] == 'meter_reading':
        yield from _yield_channel_value(
            value=channel.value,
            label=item[14:],
            metric=params['channel_type'],
            render_func=_render_meter_reading,
            levels_upper=params.get('levels_upper'),
            level_lower=params.get('levels_lower'),
        )

    elif params['channel_type'] not in _vzlogger_channel_types.keys():
        yield Result(
            state=State.UNKNOWN,
            summary=f'Unknown channel type (see details)',
            details=f'Please configure a channel type and item name in discovery rules -> vzlogger.'
        )
        yield from _yield_channel_details(channel)

    else:
        yield from _yield_channel_details(channel)
        yield Result(state=State.UNKNOWN, summary=f'not yet implemented')


register.agent_section(
    name='vzlogger',
    parse_function=parse_vzlogger,
)

register.check_plugin(
    name='vzlogger',
    service_name='%s',
    discovery_function=discovery_vzlogger,
    discovery_default_parameters={},
    discovery_ruleset_name='discovery_vzlogger',
    check_function=check_vzlogger,
    check_default_parameters={},
    check_ruleset_name='vzlogger',
)
